@@ -9,9 +9,8 @@
 use anyhow::Context as _;
 
 use std::borrow::Borrow;
-use std::collections::btree_map::{BTreeMap, Entry};
-use std::io;
 use std::path::Path;
+use std::str::FromStr;
 use std::time::SystemTime;
 
 use once_cell::sync::OnceCell;
@@ -19,18 +18,18 @@ use once_cell::sync::OnceCell;
 use sequoia_openpgp as openpgp;
 
 use openpgp::Result;
-use openpgp::{armor, Cert};
-use openpgp::Fingerprint;
+use openpgp::Cert;
 use openpgp::parse::Parse;
 use openpgp::packet::signature::subpacket::NotationData;
 use openpgp::packet::signature::subpacket::NotationDataFlags;
-use openpgp::serialize::Serialize;
 use openpgp::cert::prelude::*;
 
-use sequoia_keystore as keystore;
-
 use clap::FromArgMatches;
 
+// XXX: This could be its own crate, or preferably integrated into
+// toml_edit.
+mod toml_edit_tree;
+
 #[macro_use] mod macros;
 #[macro_use] mod log;
 
@@ -45,8 +44,11 @@ mod cli;
 use cli::SECONDS_IN_DAY;
 use cli::SECONDS_IN_YEAR;
 use cli::types::Time;
+use cli::types::Version;
+use cli::types::paths::StateDirectory;
 
 mod commands;
+pub mod config;
 pub mod output;
 pub use output::Model;
 
@@ -80,42 +82,20 @@ impl Convert<chrono::DateTime<chrono::offset::Utc>> for openpgp::types::Timestam
     }
 }
 
-/// Whether a cert or key was freshly imported, updated, or unchanged.
-///
-/// Returned by [`Sq::import_key`].
-#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
-pub enum ImportStatus {
-    /// The certificate or key is unchanged.
-    Unchanged,
-
-    /// The certificate or key is new.
-    New,
-
-    /// The certificate or key has been updated.
-    Updated,
-}
-
-impl From<keystore::ImportStatus> for ImportStatus {
-    fn from(status: keystore::ImportStatus) -> ImportStatus {
-        match status {
-            keystore::ImportStatus::Unchanged => ImportStatus::Unchanged,
-            keystore::ImportStatus::New => ImportStatus::New,
-            keystore::ImportStatus::Updated => ImportStatus::Updated,
-        }
-    }
-}
-
 /// Loads one TSK from every given file.
-fn load_keys<'a, I>(files: I) -> openpgp::Result<Vec<Cert>>
-    where I: Iterator<Item=&'a Path>
+fn load_keys<I, II>(files: I) -> openpgp::Result<Vec<Cert>>
+where
+    I: Iterator<Item=II>,
+    II: AsRef<Path>,
 {
     let mut certs = vec![];
     for f in files {
+        let f = f.as_ref();
         let cert = Cert::from_file(f)
-            .context(format!("Failed to load key from file {:?}", f))?;
+            .context(format!("Failed to load key from file {}", f.display()))?;
         if ! cert.is_tsk() {
             return Err(anyhow::anyhow!(
-                "Cert in file {:?} does not contain secret keys", f));
+                "Cert in file {} does not contain secret keys", f.display()));
         }
         certs.push(cert);
     }
@@ -123,83 +103,25 @@ fn load_keys<'a, I>(files: I) -> openpgp::Result<Vec<Cert>>
 }
 
 /// Loads one or more certs from every given file.
-fn load_certs<'a, I>(files: I) -> openpgp::Result<Vec<Cert>>
-    where I: Iterator<Item=&'a Path>
+fn load_certs<I, II>(files: I) -> openpgp::Result<Vec<Cert>>
+where
+    I: Iterator<Item=II>,
+    II: AsRef<Path>,
 {
     let mut certs = vec![];
     for f in files {
+        let f = f.as_ref();
         for maybe_cert in CertParser::from_file(f)
-            .context(format!("Failed to load certs from file {:?}", f))?
+            .context(format!("Failed to load certs from file {}", f.display()))?
         {
             certs.push(maybe_cert.context(
-                format!("A cert from file {:?} is bad", f)
+                format!("A cert from file {} is bad", f.display())
             )?);
         }
     }
     Ok(certs)
 }
 
-/// Merges duplicate certs in a keyring.
-fn merge_keyring<C>(certs: C) -> Result<BTreeMap<Fingerprint, Cert>>
-where
-    C: IntoIterator<Item = Cert>,
-{
-    let mut merged = BTreeMap::new();
-    for cert in certs {
-        match merged.entry(cert.fingerprint()) {
-            Entry::Vacant(e) => {
-                e.insert(cert);
-            },
-            Entry::Occupied(mut e) => {
-                let old = e.get().clone();
-                e.insert(old.merge_public(cert)?);
-            },
-        }
-    }
-    Ok(merged)
-}
-
-/// Serializes a keyring, adding descriptive headers if armored.
-#[allow(dead_code)]
-fn serialize_keyring(mut output: &mut dyn io::Write, certs: Vec<Cert>,
-                     binary: bool)
-                     -> openpgp::Result<()> {
-    // Handle the easy options first.  No armor no cry:
-    if binary {
-        for cert in certs {
-            cert.serialize(&mut output)?;
-        }
-        return Ok(());
-    }
-
-    // Just one Cert?  Ez:
-    if certs.len() == 1 {
-        return certs[0].armored().serialize(&mut output);
-    }
-
-    // Otherwise, merge the certs.
-    let merged = merge_keyring(certs)?;
-
-    // Then, collect the headers.
-    let mut headers = Vec::new();
-    for (i, cert) in merged.values().enumerate() {
-        headers.push(format!("Key #{}", i));
-        headers.append(&mut cert.armor_headers());
-    }
-
-    let headers: Vec<_> = headers.iter()
-        .map(|value| ("Comment", value.as_str()))
-        .collect();
-    let mut output = armor::Writer::with_headers(&mut output,
-                                                 armor::Kind::PublicKey,
-                                                 headers)?;
-    for cert in merged.values() {
-        cert.serialize(&mut output)?;
-    }
-    output.finalize()?;
-    Ok(())
-}
-
 /// Prints a warning if the user supplied "help" or "-help" to an
 /// positional argument.
 ///
@@ -213,18 +135,78 @@ fn help_warning(arg: &str) {
     }
 }
 
-// TODO: Use `derive`d command structs. No more values_of
-// TODO: Handling (and cli position) of global arguments
-fn main() -> Result<()> {
-    let mut cli = cli::build(true);
-    let matches = cli.clone().try_get_matches();
-    let c = match matches {
-        Ok(matches) => {
-            cli::SqCommand::from_arg_matches(&matches)?
+fn main() {
+    if let Err(e) = real_main() {
+        print_error_chain(&e);
+        std::process::exit(1);
+    }
+}
+
+fn real_main() -> Result<()> {
+    // In the future, we want to support multiple versions of the CLI
+    // in a single binary.  To support this, we have to parse
+    // --cli-version before parsing the arguments!  To make this
+    // unambiguous, we require that --cli-version be the very first
+    // argument, and parse it by hand.
+    let mut skip_2 = false;
+    if let Some(arg1) = std::env::args_os().nth(1) {
+        if arg1.as_encoded_bytes() == &b"--cli-version"[..] {
+            if let Some(arg2) = std::env::args_os().nth(2) {
+                let version: Version
+                    = String::from_utf8(arg2.clone().into_encoded_bytes())
+                    .map_err(Into::<anyhow::Error>::into)
+                    .and_then(|s| Ok(Version::from_str(&s)?))
+                    .with_context(|| {
+                        format!("Parsing {:?}",
+                                String::from_utf8_lossy(arg2.as_encoded_bytes()))
+                    })?;
+
+                // The version of the CLI that we implement.
+                let cli_version: Version = Version::new(
+                    usize::from_str(env!("CARGO_PKG_VERSION_MAJOR")).unwrap(),
+                    usize::from_str(env!("CARGO_PKG_VERSION_MINOR")).unwrap(),
+                    usize::from_str(env!("CARGO_PKG_VERSION_PATCH")).unwrap());
+
+                if ! version.is_acceptable_for(&cli_version) {
+                    return Err(anyhow::anyhow!(
+                        "The required CLI version, {}, is not compatible with \
+                         this version of sq, which implements version {} of \
+                         the CLI",
+                        version, cli_version));
+                }
+
+                skip_2 = true;
+            } else {
+                return Err(anyhow::anyhow!(
+                    "--cli-version missing required argument"));
+            }
         }
-        Err(err) => {
+    };
+
+    let mut cli = cli::build(true);
+    let matches = if skip_2 {
+        // Skip --cli-version and the version string.
+        let args = std::env::args_os().enumerate().filter_map(|(i, arg)| {
+            if i == 1 || i == 2 {
+                None
+            } else {
+                Some(arg)
+            }
+        });
+        cli.clone().try_get_matches_from(args)
+    } else {
+        cli.clone().try_get_matches()
+    };
+
+    let matches = match matches {
+        Ok(matches) => matches,
+        Err(mut err) => {
             // Warning: hack ahead!
             //
+            // We want to hide global options in the help output for
+            // subcommands, and we want to include values from the
+            // configuration file in the help output.
+            //
             // If we are showing the help output, we only want to
             // display the global options at the top-level; for
             // subcommands we hide the global options to not overwhelm
@@ -246,12 +228,24 @@ fn main() -> Result<()> {
             if err.kind() == ErrorKind::DisplayHelp
                 || err.kind() == ErrorKind::DisplayHelpOnMissingArgumentOrSubcommand
             {
+                // We want to try to parse the configuration file.  To
+                // that end, we first need to find the path to it.
+                let mut config = config::ConfigFile::default();
+                if let Some(augmentations) = cli::config::find_home().and_then(
+                    |home| config.read_and_augment(&home).ok())
+                {
+                    cli::config::set_augmentations(augmentations);
+                }
+
                 let output = err.render();
                 let output = if output == cli.render_long_help() {
                     Some(cli::build(false).render_long_help())
                 } else if output == cli.render_help() {
                     Some(cli::build(false).render_help())
                 } else {
+                    // Redo the parse so that the help message will
+                    // include any augmentations.
+                    err = cli::build(true).try_get_matches().unwrap_err();
                     None
                 };
 
@@ -264,17 +258,66 @@ fn main() -> Result<()> {
                     std::process::exit(err.exit_code());
                 }
             }
-            err.exit();
+
+            // Print the error message.
+            err.print()?;
+
+            // Then, figure out if this is a usage message, and
+            // extract the usage string.
+            if let Some(usage) = err.context()
+                .find_map(|(kind, value)|
+                          (kind == clap::error::ContextKind::Usage)
+                          .then_some(value))
+            {
+                print_examples(&cli, usage)?;
+            }
+
+            // Finally, exit with an error code.
+            std::process::exit(err.exit_code());
         }
     };
+    let c = cli::SqCommand::from_arg_matches(&matches)?;
+
+    let home = match &c.home {
+        Some(StateDirectory::Absolute(p)) =>
+            Some(sequoia_directories::Home::new(p.clone())?),
+        None | Some(StateDirectory::Default) =>
+            Some(sequoia_directories::Home::default()
+                 .ok_or(anyhow::anyhow!("no default SEQUOIA_HOME \
+                                         on this platform"))?
+                 .clone()),
+        Some(StateDirectory::None) => None,
+    };
+
+    // Parse the configuration file.
+    let mut config_file = config::ConfigFile::default_config(home.as_ref())?;
+    let mut config = if let Some(home) = &home {
+        // Sanity check `cli::config::find_home`.
+        debug_assert_eq!(home.location(),
+                         cli::config::find_home().unwrap().location());
+
+        config_file.read(home)
+            .with_context(|| format!(
+                "while reading configuration file {}",
+                config::ConfigFile::file_name(home).display()))?
+    } else {
+        Default::default()
+    };
 
     let time_is_now = c.time.is_none();
-    let time: SystemTime =
-        c.time.clone().unwrap_or_else(|| Time::now()).into();
+    let time: SystemTime = if let Some(t) = c.time.as_ref() {
+        t.to_system_time(std::time::SystemTime::now())?
+    } else {
+        std::time::SystemTime::now()
+    };
 
-    let mut policy = sequoia_policy_config::ConfiguredStandardPolicy::new();
-    policy.parse_default_config()?;
-    let mut policy = policy.build();
+    let policy_as_of: SystemTime = if let Some(t) = c.policy_as_of.clone() {
+        t.to_system_time(time)?
+    } else {
+        time.clone()
+    };
+
+    let mut policy = config.policy(policy_as_of)?;
 
     let known_notations_store = c.known_notation.clone();
     let known_notations = known_notations_store
@@ -283,8 +326,6 @@ fn main() -> Result<()> {
         .collect::<Vec<&str>>();
     policy.good_critical_notations(&known_notations);
 
-    let force = c.force;
-
     let mut password_cache = Vec::new();
     for password_file in &c.password_file {
         let password = std::fs::read(&password_file)
@@ -294,29 +335,89 @@ fn main() -> Result<()> {
         password_cache.push(password.into());
     };
 
+    #[allow(deprecated)]
     let sq = Sq {
+        config_file,
+        config,
         verbose: c.verbose,
-        force,
+        quiet: c.quiet,
+        overwrite: c.overwrite,
         batch: c.batch,
-        policy: &policy,
         time,
         time_is_now,
-        home: sequoia_directories::Home::new(c.home.clone())?,
-        no_rw_cert_store: c.no_cert_store,
+        policy_as_of,
+        policy: &policy,
+        home,
         cert_store_path: c.cert_store.clone(),
-        pep_cert_store_path: c.pep_cert_store.clone(),
         keyrings: c.keyring.clone(),
         keyring_tsks: Default::default(),
         cert_store: OnceCell::new(),
         trust_roots: c.trust_roots.clone(),
         trust_root_local: Default::default(),
-        no_key_store: c.no_key_store,
         key_store_path: c.key_store.clone(),
         key_store: OnceCell::new(),
         password_cache: password_cache.into(),
     };
 
-    commands::dispatch(sq, c)
+    match commands::dispatch(sq, c, &matches) {
+        Ok(()) => Ok(()),
+        Err(err) => {
+            use clap::error::ErrorFormatter;
+
+            if err.is::<clap::Error>() {
+                // We want to go from an ArgMatches to the
+                // corresponding Command.  Given an ArgMatches, we can
+                // walk down it, to get the name of each (sub)command.
+                // Then we can walk down the top-level Command to get
+                // the
+
+                // Recover the subcommands, e.g., &["key", "list"].
+                let mut m = &matches;
+                let mut trail = Vec::new();
+                while let Some((name, matches)) = m.subcommand() {
+                    trail.push(name);
+                    m = matches;
+                }
+
+                // Get the corresponding command in `cli`.
+                let mut command = Some(&mut cli);
+                for intermediate in trail.into_iter() {
+                    if let Some(c) = command {
+                        command = c.find_subcommand_mut(intermediate);
+                    }
+                }
+
+                let err = err.downcast::<clap::Error>().unwrap();
+                let exit_code = err.exit_code();
+                if let Some(mut command) = command {
+                    let err = err.format(&mut command);
+
+                    err.print()?;
+
+                    if let Some(examples) = command.get_after_help() {
+                        wprintln!("\n{}", examples);
+                    }
+                } else {
+                    // Ummm... something went wrong: we should be able
+                    // to find the subcommand.  But let's not panic.
+                    // Just show the error, and quit.
+                    if err.use_stderr() {
+                        eprintln!(
+                            "{}",
+                            clap::error::RichFormatter::format_error(&err));
+                    } else {
+                        println!(
+                            "{}",
+                            clap::error::RichFormatter::format_error(&err));
+                    }
+                }
+
+                std::process::exit(exit_code);
+            } else {
+                Err(err)
+            }
+        }
+    }
 }
 
 fn parse_notations<N>(n: N) -> Result<Vec<(bool, NotationData)>>
@@ -374,8 +475,10 @@ fn error_chain(err: &anyhow::Error) -> Vec<String> {
 
 /// Prints the error and causes, if any.
 pub fn print_error_chain(err: &anyhow::Error) {
-    wprintln!("           {}", err);
-    err.chain().skip(1).for_each(|cause| wprintln!("  because: {}", cause));
+    wprintln!();
+    wprintln!(initial_indent="  Error: ", "{}", err);
+    err.chain().skip(1).for_each(
+        |cause| wprintln!(initial_indent="because: ", "{}", cause));
 }
 
 /// Returns the error chain as a string.
@@ -424,3 +527,52 @@ where E: Borrow<anyhow::Error>,
 {
     display_error_chain(err, ": ", ", because ")
 }
+
+/// Given a `clap::Command` and a usage string, try to augment the
+/// usage message with relevant examples.
+fn print_examples(cli: &clap::Command, usage: impl ToString) -> Result<()> {
+    // First, find the invoked subcommand.
+    let usage = usage.to_string();
+    let prefix = usage.find("sq ");
+    let subcommands = if let Some(i) = prefix {
+        usage[i + 3..].split(" ")
+        // Split, but only take the parts that do not denote options
+        // or flags.
+            .take_while(|p| p.chars().all(|c| c.is_alphabetic()))
+    } else {
+        // Odd...
+        return Ok(());
+    };
+
+    // Now traverse the CLI tree to find the subcommand.
+    let mut cmd = cli;
+    for sub in subcommands {
+        cmd = if let Some(c) = cmd.get_subcommands()
+            .find(|c| c.get_name() == sub)
+        {
+            c
+        } else {
+            // Very odd...
+            return Ok(());
+        };
+    }
+
+    // And print the examples, if any.
+    if let Some(examples) = cmd.get_after_help() {
+        wprintln!("\n{}", examples);
+    }
+
+    Ok(())
+}
+
+#[cfg(test)]
+mod test {
+    use super::*;
+
+    // Run some clap tests.
+    #[test]
+    fn verify_app() {
+        let cli = cli::build(true);
+        cli.debug_assert();
+    }
+}
